# schedule.py
# SOLUTION VERSION
# Sources/people consulted: FILL IN OR WRITE "NONE"
# PUT DATE YOU COMPLETED THIS HERE
# Skeleton by Prof. Lillian Lee (LJL2), May 5, 2021

"""Defines classes Course, Schedule, StudentSchedule."""


class Schedule:
    """
    An instance represents a set of Courses.  

    An instance can represent a course roster.
    So, Courses in this Schedule can overlap in time.

    Instance attributes:

        * name [non-empty str]: the name of this Schedule, e.g., 
            "Cornell Roster Fall 2021" or (for a student schedule) "My Plan A"

        * courses [dictionary with string keys]: maps a course component name, 
            like "CS 1110 LEC 001", to a Course object  

        * warnings [list of str]: any warnings about this Schedule, such as a 
            Course time change

    """
    
    def __init__(self, n):
        """Initializes a new Schedule with name `n`, 
        attribute `courses` to the empty dictionary {},
        and attribute `warnings` set to the empty list.

        Preconditions: 
            n: non-empty string.
        """ 
        self.name = n 
        self.courses = {} 
        self.warnings = []

    def add(self, c):
        """Adds Course c to this Schedule's `courses` attribute,
        and adds this Schedule to c's list of Schedules it is in.

        Exception: if there is already a Course with the same name as `c`
        scheduled, c is NOT added. But no warning message is produced.

        Precondition: c is a Course object.
        """
        assert isinstance(c, Course),"argument is not a Course but a "+str(type(c))
        pass # STUDENTS: implement this. 
        # BEGIN REMOVE

        # act only if there is no course with this name already scheduled
        if c.name not in self.courses:
            self.courses[c.name] = c
            c.in_schedules.append(self)
        # END REMOVE

    def getCourseNames(self):
        """Returns a list of the names of the Courses in this Schedule."""
        # Don't depend on the keys of `courses` being correct
        return list(map(lambda c: c.name, self.courses.values()))

    def __str__(self):
        """Returns string representation of this Schedule."""
        out = "Schedule " + self.name + "\n" 
        for c in self.courses.values():
            out += c.name + ": " + c.getTime() + "\n"
        out += "Warnings:\n"
        for w in self.warnings:
            out += w + "\n"
        return out

    def __repr__(self):
        # Defined so that printing lists of Schedules looks better
        return self.__str__()

class StudentSchedule(Schedule):
    """
    StudentSchedules do not allow a course to be added if it conflicts
    with a course already in scheduled in it. A warning message is added when
    such an attempt is made.
    """

    # No need for separate __init__ method; just inherit the superclass's exactly.
    # No need for separate __str__ method; just inherit the superclass's exactly.

    def add(self, c):
        """Adds Course c to this Schedule (and adds this Schedule to c's list of
        Schedules).

        Exceptions: 
            * if `c` would conflict with a course already in this StudentSchedule,
              `c` is not added, and a warning message of the following form is added to
              this StudentSchedule's `warnings` attribute, for EVERY conflicting
              course: 
              "Could not add <c.name> due to conflict with <name of existing course>"

            * if there is already a Course with the same name as `c`
              scheduled, `c` is NOT added. No warning message is added, though.
        Preconditions: c is a Course object.
        """
        assert isinstance(c, Course), "argument is not a Course but a " + str(type(c))
        pass 
        # STUDENTS: Finish this implementation below this line, making effective
        # use of super().
        # Do NOT delete this comment block (so the graders can find this easily.)
        # ADD_BLOCK
        # BEGIN REMOVE

        found_conflict = False
        for existing_course_name in self.courses:
            e_c = self.courses[existing_course_name]
            if c.conflictsWith(e_c):
                found_conflict = True
                warning = "Could not add " + c.name + " due to conflict with " + \
                    e_c.name
                self.warnings.append(warning)
                # Do not return yet, because we need to add the warning for
                # each conflict.    

        if not found_conflict:
            super().add(c)

        # END REMOVE


class Course:
    """ An instance represents a class.

    Instance attributes:
        name [non-empty string]: name of the course
        in_schedules: list of Schedules this Course is in

        # STUDENTS: below this comment block (but within the docstring for class
        # Course), document any attributes you add.  
        # Don't delete this comment block, so the graders can easily see what you 
        # added.
        #
        # You'll want attribute(s) or method(s) that 
        # (1) represent the meeting time of the course,
        # (2) make it easy to retrieve the Course's times in "standard" format
        # (3) make it easy to compute whether another Course conflicts.
        # COURSE_DOCSTRING_BLOCK

        # BEGIN REMOVE
        time_string [non-empty string]: times this course meets, in "roster format"
        day_times:  dictionary, keys 'M'-'F', values are time Intervals
        # END REMOVE

        
    """
    def __init__(self, n, t):
        """
        Creates a new Course with name n, some internal representation of 
        meeting time t, and in_schedules set to the empty list

        Preconditions: 
            n: non-empty string, no other course of the same name exists.
                (Examples: "CS 1110 LEC 001", "CS 1110 DIS 201")
            t: string of the form (no beginning or trailing spaces)
                <days> <starthrs>:<startms><am or pm>-<endhrs>-<endms><am or pm>
               where <days> is a subsequence of "MTWRF"
               ***OR***
               the string "TBA"

               Examples of valid values for t:
                "MWF 10:10am-2:30pm"
                "R 8:07am-11:00am"
                "TBA"
        """
        self.name = n
        self.in_schedules = []

        # STUDENTS: implement the rest of this method below; leave the lines
        # above alone.
        # Do not delete this comment block, to help the graders find your code.
        # COURSE INIT BLOCK
        
        # BEGIN REMOVE
        self.time_string = t
        self.day_times = parse(t)
        # END REMOVE

    def getTime(self):
        """Returns a string form of this Course's time, in the format described
        by the precondition on `t` for method Course.__init__() .
        """

        return self.time_string


    def conflictsWith(self, other):
        """Returns True if this Course conflicts with `other`, False otherwise.
        It is not a conflict if one course ends when the other begins.
        Precondition: `other` is a Course object.
        """
        assert isinstance(other, Course), "argument is not a Course but a " + \
            str(type(other))

        for day in 'MTWRF':
            selfint = self.day_times[day] # None or an Interval
            otherint = other.day_times[day]
            if selfint is not None and otherint is not None:
                if selfint.overlaps(otherint):
                    return True
        # If got here, did not find any overlapping intervals           
        return False


    def changeTime(self, newtime):
        """Alters the time of this course to newtime, unless newtime is the
        same time as this Course already was.

        Also adds a string of the following form to the warnings list of every  
        Schedule this Course is listed in: 
            <name of this Course>: time changed from <oldtime> to <newtime>
        where oldtime and newtime have a format like "MWF 10:10am-2:30pm".

        Additionally, also adds a string of the following form to the warnings
        list of every StudentSchedule this Course is listed in, *if* newtime
        conflicts with any Course already in the StudentSchedule. (If there
        are multiple conflicts, each conflicting Course should be listed.)
            <name of this Course>: time change causes conflict with <name of
                conflicting course>

        Preconditions: newtime satisfies the same preconditions as `t`
        for method Course.__init__() .
        """
        assert type(newtime) == str, "Argument is not a str but a " + str(type(newtime))
        pass # STUDENTS: implement this
        # BEGIN REMOVE
        oldtime = self.time_string
        if oldtime == newtime:
            return # Do nothing

        self.day_times = parse(newtime)
        self.time_string = newtime
        for s in self.in_schedules:
            s.warnings.append(self.name + ": time changed from " + oldtime + \
                                " to " + newtime)
            if isinstance(s, StudentSchedule):
                for c in s.courses.values():
                    if c != self and c.conflictsWith(self):
                        s.warnings.append(self.name + ": time change causes" + \
                            " conflict with " + c.name)


        # END REMOVE

    def getScheduleNames(self):
        """Returns a list of the names of the Schedules this Course is in."""
        # Don't depend on the keys of `courses` being correct
        return list(map(lambda s: s.name, self.in_schedules))

    def __str__(self):
        """Returns a two-line string of the form (so includes "\n" in the middle)
            <Course name>: <Meeting Pattern>
            In schedules: <name of sched1>, <name of sched2>, ...
        where <Meeting Pattern> obeys the same formatting conditions as `t` for
        method Course.__init__(). No trailing spaces and the string should not
        end in a comma.

        Exception: If this Course is not in any Schedules, the string returned is: 
            <Course name>: <Meeting Pattern>
        """
        return self.name + ": " + self.getTime() + "\nIn schedules: " + \
            ", ".join(self.getScheduleNames())

    def __repr__(self):
        # Defined so that printing lists of courses looks better
        return self.__str__()

# STUDENTS: add any helper functions below this line. 
# Do not delete this comment block.
# HELPER FUNCTIONS  BLOCK

# BEGIN REMOVE
def parse(t):
    """Returns a 'MTWRF' dictionary based on t, a "roster-format" description of
    course meeting times, including TBA"""

    out = {}
    for day in 'MTWRF':
        out[day] = None

    if t.strip() == "TBA":
        return out

    # Example: MWF 9:05am-9:55pm
    temp1 = t.split()
    tdays = temp1[0].strip() # "MWF"
    temp2 = temp1[1].split('-')
    tstart = temp2[0] #9:05am
    tend = temp2[1].strip() # 9:55pm
    start = convert_to_minutes(tstart)
    end = convert_to_minutes(tend)

    for day in tdays:
        out[day] = Interval(start, end)

    return out

def convert_to_minutes(s):
    """Returns the minutes-from-start-of-day that string s represents.
    Precondition: s is a string like "12:01am" or "3:00pm"."""

    s = s.strip() # Just in case.
    start=0
    if s.endswith('pm'):
        start = 12*60
    parts = s.split(':')
    # Thanks to Brynn Szczesniak and Elisabeth Finkel for catching this
    if parts[0] == "12":
        parts[0] = "0"
    hrs = int(parts[0])*60
    mins = int(parts[1][:-2])

    return start + hrs + mins



# END REMOVE

# STUDENTS: add any new classes below this line.
# Do not delete this comment block.
# NEW CLASSES BLOCK

# BEGIN REMOVE
class Interval:
    """Attributes: start [int] and end [int], holding minutes from the start
    of a day."""

    def __init__(self, s, e):
        """New Interval has start s and end e"""
        self.start = s
        self.end = e

    def overlaps(self, other_ivl):
        """Returns True if this Interval starts within other_ivl or ends within
        other_ivl, where other_ivl is an Interval object."""
        return other_ivl.start <= self.start < other_ivl.end or \
            other_ivl.start < self.end <= other_ivl.end or \
            self.start <= other_ivl.start < self.end or \
            self.start < other_ivl.end <= self.end


    def __str__(self):
        return "(" + str(self.start) + ", " + str(self.end) + ")"

# END REMOVE


# STUDENTS: this is some quick-and-dirty testing code I wrote up along the way.
# It is meant to inspire you, should you need to make such tests yourself.
# if __name__ == '__main__':
#     import roster_creation
#     initial_courses_to_add = roster_creation.make_initial_course_list()
#     roster = roster_creation.make_sample_roster(initial_courses_to_add)
#     arch1120 = roster.courses["ARCH 1120 STU 520"
#     my_schedule = StudentSchedule("My Plan C")
#     test_courses = [roster.courses["CS 1110 LEC 001"],
#                     roster.courses["AEP 3560 LEC 001"],
#                     roster.courses["FAKER 0000 LEC 002"],
#                     arch1120]             
#     for c in test_courses:
#         my_schedule.add(c)

#     arch1120.changeTime("MWF 8:00am-8:00pm")
#     print(arch1120)
#     aep = roster.courses["AEP 3560 LEC 001"]
#
#     < started doing prints of various subcomponents of my Courses to check them>
#     print(my_schedule.warnings)
#     # If got here, did not find any overlapping intervals           
#     print("outside")
